@@ -22,7 +22,8 @@ class HTTPClient(base.HTTPClient):
                                          **connector_kwargs)
         session_kwargs = {}
         if connections_timeout:
-            session_kwargs['timeout'] = connections_timeout
+            timeout = aiohttp.ClientTimeout(total=connections_timeout)
+            session_kwargs['timeout'] = timeout
         self._session = aiohttp.ClientSession(connector=connector,
                                               **session_kwargs)
 
@@ -63,7 +64,7 @@ class Consul(base.Consul):
         self.connections_timeout = connections_timeout
         super().__init__(*args, **kwargs)
 
-    def connect(self, host, port, scheme, verify=True, cert=None):
+    def http_connect(self, host, port, scheme, verify=True, cert=None):
         return HTTPClient(host, port, scheme, loop=self._loop,
                           connections_limit=self.connections_limit,
                           connections_timeout=self.connections_timeout,
