@@ -48,7 +48,7 @@ class Check:
     There are three different kinds of checks: script, http and ttl
     """
     @classmethod
-    def script(klass, args, interval):
+    def script(klass, args, interval, deregister=None):
         """
         Run the script *args* every *interval* (e.g. "10s") to peform health
         check
@@ -57,7 +57,10 @@ class Check:
             warnings.warn(
                 "Check.script should take a list of args", DeprecationWarning)
             args = ["sh", "-c", args]
-        return {'args': args, 'interval': interval}
+        ret = {'args': args, 'interval': interval}
+        if deregister:
+            ret['DeregisterCriticalServiceAfter'] = deregister
+        return ret
 
     @classmethod
     def http(klass, url, interval, timeout=None, deregister=None, header=None):
@@ -274,6 +277,10 @@ class HTTPClient(metaclass=abc.ABCMeta):
     def post(self, callback, path, params=None, data=''):
         raise NotImplementedError
 
+    @abc.abstractmethod
+    def close(self):
+        raise NotImplementedError
+
 
 class Consul:
     def __init__(
@@ -320,7 +327,7 @@ class Consul:
         if os.getenv('CONSUL_HTTP_SSL_VERIFY') is not None:
             verify = os.getenv('CONSUL_HTTP_SSL_VERIFY') == 'true'
 
-        self.http = self.connect(host, port, scheme, verify, cert)
+        self.http = self.http_connect(host, port, scheme, verify, cert)
         self.token = os.getenv('CONSUL_HTTP_TOKEN', token)
         self.scheme = scheme
         self.dc = dc
@@ -340,6 +347,19 @@ class Consul:
         self.query = Consul.Query(self)
         self.coordinate = Consul.Coordinate(self)
         self.operator = Consul.Operator(self)
+        self.connect = Consul.Connect(self)
+
+    def __enter__(self):
+        return self
+
+    async def __aenter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.http.close()
+
+    async def __aexit__(self, exc_type, exc, tb):
+        await self.http.close()
 
     class Event:
         """
@@ -697,6 +717,7 @@ class Consul:
             self.agent = agent
             self.service = Consul.Agent.Service(agent)
             self.check = Consul.Agent.Check(agent)
+            self.connect = Consul.Agent.Connect(agent)
 
         def self(self):
             """
@@ -717,6 +738,15 @@ class Consul:
             """
             return self.agent.http.get(CB.json(), '/v1/agent/services')
 
+        def service_definition(self, service_id):
+            """
+            Returns a service definition for a single instance that is registered
+            with the local agent.
+            """
+            return self.agent.http.get(
+                CB.json(),
+                '/v1/agent/service/%s' % service_id)
+
         def checks(self):
             """
             Returns all the checks that are registered with the local agent.
@@ -746,7 +776,7 @@ class Consul:
             return self.agent.http.get(
                 CB.json(), '/v1/agent/members', params=params)
 
-        def maintenance(self, enable, reason=None):
+        def maintenance(self, enable, reason=None, token=None):
             """
             The node maintenance endpoint can place the agent into
             "maintenance mode".
@@ -763,11 +793,14 @@ class Consul:
             params.append(('enable', enable))
             if reason:
                 params.append(('reason', reason))
+            token = token or self.agent.token
+            if token:
+                params.append(('token', token))
 
             return self.agent.http.put(
                 CB.bool(), '/v1/agent/maintenance', params=params)
 
-        def join(self, address, wan=False):
+        def join(self, address, wan=False, token=None):
             """
             This endpoint instructs the agent to attempt to connect to a
             given address.
@@ -783,6 +816,9 @@ class Consul:
 
             if wan:
                 params.append(('wan', 1))
+            token = token or self.agent.token
+            if token:
+                params.append(('token', token))
 
             return self.agent.http.put(
                 CB.bool(), '/v1/agent/join/%s' % address, params=params)
@@ -903,16 +939,22 @@ class Consul:
                     params=params,
                     data=json.dumps(payload))
 
-            def deregister(self, service_id):
+            def deregister(self, service_id, token=None):
                 """
                 Used to remove a service from the local agent. The agent will
                 take care of deregistering the service with the Catalog. If
                 there is an associated check, that is also deregistered.
                 """
-                return self.agent.http.put(
-                    CB.bool(), '/v1/agent/service/deregister/%s' % service_id)
+                params = []
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
 
-            def maintenance(self, service_id, enable, reason=None):
+                return self.agent.http.put(
+                    CB.bool(), '/v1/agent/service/deregister/%s' % service_id,
+                    params=params)
+
+            def maintenance(self, service_id, enable, reason=None, token=None):
                 """
                 The service maintenance endpoint allows placing a given service
                 into "maintenance mode".
@@ -933,6 +975,10 @@ class Consul:
                 if reason:
                     params.append(('reason', reason))
 
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
+
                 return self.agent.http.put(
                     CB.bool(),
                     '/v1/agent/service/maintenance/{0}'.format(service_id),
@@ -1018,15 +1064,21 @@ class Consul:
                     params=params,
                     data=json.dumps(payload))
 
-            def deregister(self, check_id):
+            def deregister(self, check_id, token=None):
                 """
                 Remove a check from the local agent.
                 """
+                params = []
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
+
                 return self.agent.http.put(
                     CB.bool(),
-                    '/v1/agent/check/deregister/%s' % check_id)
+                    '/v1/agent/check/deregister/%s' % check_id,
+                    params=params)
 
-            def ttl_pass(self, check_id, notes=None):
+            def ttl_pass(self, check_id, notes=None, token=None):
                 """
                 Mark a ttl based check as passing. Optional notes can be
                 attached to describe the status of the check.
@@ -1034,13 +1086,16 @@ class Consul:
                 params = []
                 if notes:
                     params.append(('note', notes))
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
 
                 return self.agent.http.put(
                     CB.bool(),
                     '/v1/agent/check/pass/%s' % check_id,
                     params=params)
 
-            def ttl_fail(self, check_id, notes=None):
+            def ttl_fail(self, check_id, notes=None, token=None):
                 """
                 Mark a ttl based check as failing. Optional notes can be
                 attached to describe why check is failing. The status of the
@@ -1049,13 +1104,16 @@ class Consul:
                 params = []
                 if notes:
                     params.append(('note', notes))
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
 
                 return self.agent.http.put(
                     CB.bool(),
                     '/v1/agent/check/fail/%s' % check_id,
                     params=params)
 
-            def ttl_warn(self, check_id, notes=None):
+            def ttl_warn(self, check_id, notes=None, token=None):
                 """
                 Mark a ttl based check with warning. Optional notes can be
                 attached to describe the warning. The status of the
@@ -1064,12 +1122,78 @@ class Consul:
                 params = []
                 if notes:
                     params.append(('note', notes))
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
 
                 return self.agent.http.put(
                     CB.bool(),
                     '/v1/agent/check/warn/%s' % check_id,
                     params=params)
 
+        class Connect:
+            def __init__(self, agent):
+                self.agent = agent
+                self.ca = Consul.Agent.Connect.CA(agent)
+
+            def authorize(
+                    self,
+                    target,
+                    client_cert_uri,
+                    client_cert_serial,
+                    token=None):
+                """
+                Tests whether a connection attempt is authorized between
+                two services.
+                More information is available
+                `here <https://www.consul.io/api-docs/agent/connect>`_.
+
+                *target* is the name of the service that is being requested.
+
+                *client_cert_uri* The unique identifier for the requesting
+                client.
+
+                *client_cert_serial* The colon-hex-encoded serial number for
+                the requesting client cert.
+                """
+
+                payload = {
+                    'Target': target,
+                    'ClientCertURI': client_cert_uri,
+                    'ClientCertSerial': client_cert_serial
+                }
+
+                params = []
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
+
+                return self.agent.http.put(
+                    CB.bool(),
+                    '/v1/agent/connect/authorize',
+                    params=params,
+                    data=json.dumps(payload))
+
+            class CA:
+                def __init__(self, agent):
+                    self.agent = agent
+
+                def roots(self):
+                    return self.agent.http.get(
+                        CB.json(),
+                        '/v1/agent/connect/ca/roots')
+
+                def leaf(self, service, token=None):
+                    params = []
+                    token = token or self.agent.token
+                    if token:
+                        params.append(('token', token))
+
+                    return self.agent.http.get(
+                        CB.bool(),
+                        '/v1/agent/connect/ca/leaf/%s' % service,
+                        params=params)
+
     class Catalog:
         def __init__(self, agent):
             self.agent = agent
@@ -1409,17 +1533,47 @@ class Consul:
                 '/v1/catalog/node/%s' % node,
                 params=params)
 
+        def _service(self,
+                     internal_uri,
+                     index=None,
+                     wait=None,
+                     tag=None,
+                     consistency=None,
+                     dc=None,
+                     near=None,
+                     token=None,
+                     node_meta=None):
+            params = []
+            dc = dc or self.agent.dc
+            if dc:
+                params.append(('dc', dc))
+            if tag:
+                params.append(('tag', tag))
+            if index:
+                params.append(('index', index))
+                if wait:
+                    params.append(('wait', wait))
+            if near:
+                params.append(('near', near))
+            token = token or self.agent.token
+            if token:
+                params.append(('token', token))
+            consistency = consistency or self.agent.consistency
+            if consistency in ('consistent', 'stale'):
+                params.append((consistency, '1'))
+            if node_meta:
+                for nodemeta_name, nodemeta_value in node_meta.items():
+                    params.append(('node-meta', '{0}:{1}'.
+                                   format(nodemeta_name, nodemeta_value)))
+            return self.agent.http.get(
+                CB.json(index=True),
+                internal_uri,
+                params=params)
+
         def service(
                 self,
                 service,
-                index=None,
-                wait=None,
-                tag=None,
-                consistency=None,
-                dc=None,
-                near=None,
-                token=None,
-                node_meta=None):
+                **kwargs):
             """
             Returns a tuple of (*index*, *nodes*) of the nodes providing
             *service* in the *dc* datacenter. *dc* defaults to the current
@@ -1460,76 +1614,38 @@ class Consul:
                     }
                 ])
             """
-            params = []
-            dc = dc or self.agent.dc
-            if dc:
-                params.append(('dc', dc))
-            if tag:
-                params.append(('tag', tag))
-            if index:
-                params.append(('index', index))
-                if wait:
-                    params.append(('wait', wait))
-            if near:
-                params.append(('near', near))
-            token = token or self.agent.token
-            if token:
-                params.append(('token', token))
-            consistency = consistency or self.agent.consistency
-            if consistency in ('consistent', 'stale'):
-                params.append((consistency, '1'))
-            if node_meta:
-                for nodemeta_name, nodemeta_value in node_meta.items():
-                    params.append(('node-meta', '{0}:{1}'.
-                                   format(nodemeta_name, nodemeta_value)))
-            return self.agent.http.get(
-                CB.json(index=True),
-                '/v1/catalog/service/%s' % service,
-                params=params)
+            internal_uri = '/v1/catalog/service/%s' % service
+            return self._service(internal_uri=internal_uri, **kwargs)
+
+        def connect(
+                self,
+                service,
+                **kwargs):
+            """
+            Returns a tuple of (*index*, *nodes*) of the nodes providing
+            connect-capable *service* in the *dc* datacenter. *dc* defaults
+            to the current datacenter of this agent.
+
+            Request arguments and response format are the same as catalog.service
+            """
+            internal_uri = '/v1/catalog/connect/%s' % service
+            return self._service(internal_uri=internal_uri, **kwargs)
 
     class Health:
         # TODO: All of the health endpoints support all consistency modes
         def __init__(self, agent):
             self.agent = agent
 
-        def service(self,
-                    service,
-                    index=None,
-                    wait=None,
-                    passing=None,
-                    tag=None,
-                    dc=None,
-                    near=None,
-                    token=None,
-                    node_meta=None):
-            """
-            Returns a tuple of (*index*, *nodes*)
-
-            *index* is the current Consul index, suitable for making subsequent
-            calls to wait for changes since this query was last run.
-
-            *wait* the maximum duration to wait (e.g. '10s') to retrieve
-            a given index. this parameter is only applied if *index* is also
-            specified. the wait time by default is 5 minutes.
-
-            *nodes* are the nodes providing the given service.
-
-            Calling with *passing* set to True will filter results to only
-            those nodes whose checks are currently passing.
-
-            Calling with *tag* will filter the results by tag.
-
-            *dc* is the datacenter of the node and defaults to this agents
-            datacenter.
-
-            *near* is a node name to sort the resulting list in ascending
-            order based on the estimated round trip time from that node
-
-            *token* is an optional `ACL token`_ to apply to this request.
-
-            *node_meta* is an optional meta data used for filtering, a
-            dictionary formatted as {k1:v1, k2:v2}.
-            """
+        def _service(self,
+                     internal_uri,
+                     index=None,
+                     wait=None,
+                     passing=None,
+                     tag=None,
+                     dc=None,
+                     near=None,
+                     token=None,
+                     node_meta=None):
             params = []
             if index:
                 params.append(('index', index))
@@ -1538,7 +1654,10 @@ class Consul:
             if passing:
                 params.append(('passing', '1'))
             if tag is not None:
-                params.append(('tag', tag))
+                if not isinstance(tag, list):
+                    tag = [tag]
+                for tag_item in tag:
+                    params.append(('tag', tag_item))                
             dc = dc or self.agent.dc
             if dc:
                 params.append(('dc', dc))
@@ -1553,9 +1672,53 @@ class Consul:
                                    format(nodemeta_name, nodemeta_value)))
             return self.agent.http.get(
                 CB.json(index=True),
-                '/v1/health/service/%s' % service,
+                internal_uri,
                 params=params)
 
+        def service(self, service, **kwargs):
+            """
+           Returns a tuple of (*index*, *nodes*)
+
+           *index* is the current Consul index, suitable for making subsequent
+           calls to wait for changes since this query was last run.
+
+           *wait* the maximum duration to wait (e.g. '10s') to retrieve
+           a given index. this parameter is only applied if *index* is also
+           specified. the wait time by default is 5 minutes.
+
+           *nodes* are the nodes providing the given service.
+
+           Calling with *passing* set to True will filter results to only
+           those nodes whose checks are currently passing.
+
+           Calling with *tag* will filter the results by tag, multiple tags
+           using list possible.
+
+           *dc* is the datacenter of the node and defaults to this agents
+           datacenter.
+
+           *near* is a node name to sort the resulting list in ascending
+           order based on the estimated round trip time from that node
+
+           *token* is an optional `ACL token`_ to apply to this request.
+
+           *node_meta* is an optional meta data used for filtering, a
+           dictionary formatted as {k1:v1, k2:v2}.
+           """
+            internal_uri = '/v1/health/service/%s' % service
+            return self._service(internal_uri=internal_uri, **kwargs)
+
+        def connect(self, service, **kwargs):
+            """
+            Returns a tuple of (*index*, *nodes*) of the nodes providing
+            connect-capable *service* in the *dc* datacenter. *dc* defaults
+            to the current datacenter of this agent.
+
+            Request arguments and response format are the same as health.service
+            """
+            internal_uri = '/v1/health/connect/%s' % service
+            return self._service(internal_uri=internal_uri, **kwargs)
+
         def checks(
                 self,
                 service,
@@ -2436,3 +2599,35 @@ class Consul:
             """
             return self.agent.http.get(
                 CB.json(), '/v1/operator/raft/configuration')
+
+    class Connect:
+        def __init__(self, agent):
+            self.agent = agent
+            self.ca = Consul.Connect.CA(agent)
+
+        class CA:
+            def __init__(self, agent):
+                self.agent = agent
+
+            def roots(self, pem=False, token=None):
+                params = []
+                params.append(('pem', int(pem)))
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
+
+                return self.agent.http.get(
+                    CB.bool(),
+                    '/v1/connect/ca/roots',
+                    params=params)
+
+            def configuration(self, token=None):
+                params = []
+                token = token or self.agent.token
+                if token:
+                    params.append(('token', token))
+
+                return self.agent.http.get(
+                    CB.bool(),
+                    '/v1/connect/ca/configuration',
+                    params=params)
